#!/usr/bin/python3
import random
import itertools
import sys
from decimal import *

if len(sys.argv) != 5:
    print("usage: gen inp_mult inp_dir GEN desc_file\n", file=sys.stderr)
    sys.exit(0)

sys.setrecursionlimit(10000)

getcontext().prec = 200

def randbits(k):
    return random.getrandbits(k)

class HistoryCoin:
    def __init__(self, n):
        self.reset_prob = 3.0/n
        self.l =  1.0/random.gammavariate(0.4, 2.0)
        self.sep = 0.5
    def flip(self):
        if(random.random() < self.reset_prob):
            self.l =  1.0/random.gammavariate(0.4, 2.0)
        self.sep += random.choice((-1.0, 1.0))*random.expovariate(self.l)
        self.sep %= 1.0
        if(random.random() < self.sep):
            return 0
        return 1


ncr_lookup = [[1]]
for i in range(1,1025):
    ncr_lookup.append([1])
    for j in range(1,i):
        ncr_lookup[-1].append(ncr_lookup[i-1][j-1]+ncr_lookup[i-1][j])
    ncr_lookup[-1].append(1)

def nCr(n, r):
    return ncr_lookup[n][r]

def transitions(b, act, m):
    remain = 2**m
    res = []
    for i in range(act, b+1):
        combs = nCr(b-act, i-act)
        cur = min(remain, combs)
        remain -= cur
        res.append((i, Decimal(cur)/Decimal(2**m)))
        if remain == 0:
            break
    if remain != 0:
        res.append((-1, Decimal(remain)/Decimal(2**m)))
    return res

# b, m
test_group = [(2**i, 2**j) for i in range(4, 10) for j in range(3, i) ]
test_group.remove((2**9, 2**3))
print(len(test_group))
print(test_group)

bn_lim = 120*1000
tests_in_group = int(sys.argv[1])
inp_dir = sys.argv[2]
GEN_file = open(sys.argv[3], "w")
desc_file = open(sys.argv[4], "w")
LD_PRECISION = Decimal("0."+20*"0"+"1")


test_idx = 0
for i in range(len(test_group)):
    GEN_file.write("# ST: 5\n")
    b, m = test_group[i]
    desc_str = str(test_idx)+"-"+str(test_idx+tests_in_group-1)+": b="+str(b)+", m="+str(m)
    desc_file.write(desc_str+"\n")
    print("\n"+desc_str)

    nxt_state = [ transitions(b, j, m) for j in range(b+1)]
    exp_moves = Decimal(0)
    lim_prob = Decimal("0."+50*"9")
    cum_lim_prob = Decimal("0."+19*"0"+"1")
    end_prob = Decimal(0)
    act_prob = [Decimal(0) for _ in range(b+1)]
    act_prob[0] = Decimal(1)
    move_lim = 0
    end_probs = [end_prob]
    while(end_prob < lim_prob):
        # print(move_lim)
        # print(end_prob)
        nxt_act_prob = [Decimal(0) for _ in range(b+1)]
        for j in range(b+1):
            for nxt, prob in nxt_state[j]:
                prob *= act_prob[j]
                if nxt==-1:
                    end_prob += prob
                    exp_moves += prob * Decimal(move_lim)
                else:
                    nxt_act_prob[nxt] += prob
        act_prob = nxt_act_prob
        end_probs.append(end_prob)
        move_lim += 1
    print(move_lim)
    print(end_prob)

    cum_move_lim = move_lim
    reinit_cnt = 0
    for i in itertools.count(1):
        while(cum_move_lim-1 > exp_moves and (Decimal("1")-end_probs[cum_move_lim-1])**i < cum_lim_prob):
            cum_move_lim -= 1
        if(i*cum_move_lim*b > bn_lim):
            break
        reinit_cnt += 1
    print(reinit_cnt)

    for j in range(tests_in_group):
        GEN_file.write(str(test_idx)+"\n")
        fout = open(inp_dir+"/input"+str(test_idx)+".txt", "w")
        fout.write(str(b)+" "+str(m)+"\n")
        fout.write(str(exp_moves.quantize(LD_PRECISION))+"\n")
        fout.write(str(reinit_cnt)+" "+str(move_lim)+"\n")

        encode_coin = HistoryCoin(int(exp_moves))
        decode_coin = HistoryCoin(int(exp_moves))
        for k in range(reinit_cnt):
            for k in range(move_lim):
                fout.write(format(randbits(m), "b").zfill(m)+" "+str(encode_coin.flip())+" "+str(decode_coin.flip())+"\n")
        test_idx += 1
        fout.close()
GEN_file.close()
desc_file.close()
